// import 'dart:math';

import 'dart:io';
import 'dart:math';

import 'package:kenken/kenken.dart';
import 'package:test/test.dart';

void main() {
  group('Functions:', () {
    test('isUniq works right', () {
      expect(isUniq([1, 2, 3]), equals(true));
      expect(isUniq([1, 2, 1]), equals(false));
      expect(isUniq([1, 2, 2]), equals(false));
      expect(
          isUniq([
            1,
            2,
          ]),
          equals(true));
      expect(
          isUniq([
            1,
          ]),
          equals(true));
    });
  });

  group('Rule object', () {
    test('return cell to display rule in it', () {
      // Let create rule "5+" with such cells:
      // |-----------------------|
      // |      |        |       | # r - row
      // |------=================| # c - column
      // |      || r2,c2 | r2,c3 |
      // |---------------========|
      // |      || r3,c2 ||      |
      // |------==========-------|
      // In which cell to display rule "5+" ?
      var rule =
          Rule([RC(2, 2), RC(2, 3), RC(3, 2)], Operation(sum, label: '+'), 5);
      // Right answer is RC(2,2), because it is topmost and leftmost:
      var cellToDisplay = rule.coordinatesToDisplayTheRule;
      expect(cellToDisplay.row, equals(2));
      expect(cellToDisplay.col, equals(2));
    });
  });

  group('Kenken (filled manually):', () {
    Kenken kenken;

    setUp(() {
      kenken = Kenken(3);

      kenken
        // RC 1,1 shoul be equal 1:
        ..addRuleFor([RC(1, 1)], Operation(singleValue, label: '1'), 1)
        // Sum of cells 1,2 and 1,3 should be equal 5:
        ..addRuleFor([RC(1, 2), RC(1, 3)], Operation(sum, label: '+'), 5)
        // Diff between 2,1 and 3,1 should be 1
        // Note: Using kenken.RCList() instead of RC()
        ..addRuleFor(
            kenken.RCList([
              [2, 1],
              [3, 1]
            ]),
            Operation(diff, label: '-'),
            1)
        // sum of 2,2 2,3 should be 4
        ..addRuleFor([RC(2, 2), RC(2, 3)], Operation(sum, label: '+'), 4);

      // Do not add rule below to kenken now.
      // Let kenken will be not fully covered by custom reles.
      // This rule we will add later in test.
      // sum of 3,2 and 3,3 should be 3:
      // ..addRuleFor([RC(3, 2), RC(3, 3)], Operation(sum, label: '+'), 3);
    });

    test('`coordinatesOfRow` method returns right coordinates', () {
      var row2Coordinates = kenken.coordinatesOfRow(2);
      // first coordinate of row2Coordinates:
      expect(row2Coordinates[0].row, equals(2));
      expect(row2Coordinates[0].col, equals(1));
      // third coordinate of row2Coordinates:
      expect(row2Coordinates[2].row, equals(2));
      expect(row2Coordinates[2].col, equals(3));
    });

    test('rules not cover full kenken', () {
      expect(kenken.notCoveredByCustomRulesCells.length, equals(2));
    });

    test('rules cover full kenken', () {
      // Let add final rule.
      // sum of 3,2 and 3,3 should be 3:
      // Now the kenken is fully covered by custom rules.
      kenken.addRuleFor([RC(3, 2), RC(3, 3)], Operation(sum, label: '+'), 3);
      expect(kenken.notCoveredByCustomRulesCells.length, equals(0));
    });

    test('Right solved kenken has not invalid rules', () {
      var values = [
        [1, 2, 3], // row = 1,
        [2, 3, 1], // row = 2,
        [3, 1, 2], // row = 3,
      ];

      // Fill kenken cells with values as it does user:
      for (var r in kenken.rowOrColNumbers) {
        for (var c in kenken.rowOrColNumbers) {
          kenken.setValue(r, c, values[r - 1][c - 1]);
        }
      }

      // expect(awesome.isAwesome, isTrue);
      expect(kenken.brokenRules, equals([]));

      print('$kenken');
    });

    // вынести в другой кенкен, незаполненный кастомными правилами
    // test('Kenken auto fill values',(){
    //   var autoFilled = kenken.autoGeneratedValues;

    //   for (var r in kenken.rowOrColNumbers){
    //     for (var c in kenken.rowOrColNumbers){
    //       kenken.setValue(r, c, autoFilled[r-1][c-1]);
    //     }
    //   }
    //   // expect(awesome.isAwesome, isTrue);
    //   expect(kenken.brokenRules, equals([]));

    //   print('$kenken');

    //   // print('auto filled values: $autoFilled');

    // });
  });

  group('Kenken (filled automatically):', () {
    Kenken kenken;

    setUp(() {
      kenken = Kenken(3);
    });

    test('Kenken auto fill values', () {
      var autoFilled = kenken.autoGeneratedValues;
      // print('auto filled values: $autoFilled');

      for (var r in kenken.rowOrColNumbers) {
        for (var c in kenken.rowOrColNumbers) {
          kenken.setValue(r, c, autoFilled[r - 1][c - 1]);
        }
      }

      expect(kenken.brokenRules, equals([]));
      // print('$kenken');

      // var randomRCList = <RC>[];
      // randomRCList.add(kenken.randomRC);
      // var la = randomRCList.last;
      // // td: можно в цикле нужное кол-во раз пытаться найти группу ячеек длиной N
      // var cellNear = kenken.rcNearWithoutRule(la.row, la.col);
      // if (cellNear != null) {
      //   randomRCList.add(cellNear);
      // }
    });

    test('Kenken auto fill values', () {
      // var autoFilled = kenken.autoGeneratedValues;
      // // print('auto filled values: $autoFilled');

      // for (var r in kenken.rowOrColNumbers) {
      //   for (var c in kenken.rowOrColNumbers) {
      //     kenken.setValue(r, c, autoFilled[r - 1][c - 1]);
      //   }
      // }

      // kenken.autoFillValuesWithoutRules();

      // var listOfList = <List<RC>>[];
      // var randomRC = kenken.randomRC;
      // var list = kenken.tryExpandRC(randomRC, 3);
      // var values = list.map( (rc)=>kenken.valueOf(rc)).toList();
      // kenken.addRuleFor(list, Operation(sum,label: '+'), sum(values));

      // listOfList.add(list);
      // var lastRC = listOfList.last.last;

      // var near2 = kenken.notCoveredByCustomRulesCells.first;
      // // var near2 = kenken.rcNearWithoutRule(lastRC.row, lastRC.col);
      // // if
      // var list2 = kenken.tryExpandRC(near2, 3);
      // var values2 = list2.map( (rc)=>kenken.valueOf(rc)).toList();
      // kenken.addRuleFor(list2, Operation(sum,label: '+'), sum(values2));
      // listOfList.add(list2);

      // var near3 = kenken.notCoveredByCustomRulesCells.first;
      // var list3 = kenken.tryExpandRC(near3, 3);
      // var values3 = list3.map( (rc)=>kenken.valueOf(rc)).toList();
      // kenken.addRuleFor(list3, Operation(sum,label: '+'), sum(values3));
      // listOfList.add(list3);

      // var randomRC = kenken.randomRC;
      // var list = kenken.tryExpandRC(randomRC, 3);
      // var values = list.map((rc) => kenken.valueOf(rc)).toList();
      // kenken.addRuleFor(list, Operation(sum, label: '+'), sum(values));
      // while (kenken.notCoveredByCustomRulesCells.isNotEmpty) {
      //   var startRC =
      //       kenken.notCoveredByCustomRulesCells.length == kenken.size ^ 2
      //           ? kenken.randomRC
      //           : kenken.notCoveredByCustomRulesCells.first;
      //   ;
      //   var wantedSize = 3;
      //   var rcList = kenken.tryExpandRC(startRC, wantedSize);
      //   var rcListValues = rcList.map((rc) => kenken.valueOf(rc)).toList();
      //   var operation = Operation(sum,label: '+');
      //   var validOperationResult = sum(rcListValues);
      //   kenken.addRuleFor(rcList, operation, validOperationResult);
      // }

      kenken.autoFillCustomRules();

      // print(listOfList);
      print(kenken);
      print(kenken.customRules);
    });
  });
}
